package de.tudresden.inf.st.ros2rag.receiverstub;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.dataformat.yaml.YAMLFactory;
import com.google.protobuf.InvalidProtocolBufferException;
import config.Dataconfig.DataConfig;
import config.Robotconfig.RobotConfig;
import de.tudresden.inf.st.ros2rag.starter.ast.MqttHandler;
import de.tudresden.inf.st.ros2rag.common.Util;
import de.tudresden.inf.st.ros2rag.common.DataConfiguration;
import de.tudresden.inf.st.ros2rag.common.DataConfiguration.ActualConfiguration;
import org.apache.logging.log4j.Level;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import panda.Linkstate.PandaLinkState;
import plan.TrajectoryOuterClass;
import plan.TrajectoryOuterClass.Trajectory;

import java.io.File;
import java.io.IOException;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

public class ReceiverMain {

  private static final Logger logger = LogManager.getLogger(ReceiverMain.class);
  private String topicPattern;

  public static void main(String[] args) throws Exception {
    ReceiverMain main = new ReceiverMain();
    ObjectMapper mapper = new ObjectMapper(new YAMLFactory());
    File configFile = new File(args[0]);
    System.out.println("Using config file: " + configFile.getAbsolutePath());
    ActualConfiguration config = mapper.readValue(configFile, DataConfiguration.class).panda_mqtt_connector;
    main.run(config);
  }

  private void run(ActualConfiguration config) throws IOException, InterruptedException {
    final CountDownLatch finish = new CountDownLatch(1);

    AtomicInteger topicMaxLength = new AtomicInteger();
    MqttHandler receiver = new MqttHandler("receiver stub");
    Util.setMqttHost(receiver::setHost, config);
    receiver.waitUntilReady(2, TimeUnit.SECONDS);
    receiver.newConnection(config.topics.robotConfig, this::printRobotConfig);
    receiver.newConnection(config.topics.dataConfig, this::printDataConfig);
    receiver.newConnection(config.topics.trajectory, this::printTrajectory);
    receiver.newConnection(config.topics.nextStep, this::printNextStep);

    Util.iterateLinks((isEndEffector, topic, name) -> {
      receiver.newConnection(topic, this::printPandaLinkState);
      topicMaxLength.set(Math.max(topicMaxLength.get(), topic.length()));
    }, config);
    this.topicPattern = "%" + topicMaxLength + "s";

    receiver.newConnection("components", bytes -> {
      String message = new String(bytes);
      logger.info("Components: {}", message);
    });

    receiver.newConnection("receiver", bytes -> {
      String message = new String(bytes);
      if (message.equals("exit")) {
        finish.countDown();
      }
    });

    Runtime.getRuntime().addShutdownHook(new Thread(receiver::close));
    finish.await();
    receiver.close();
  }

  private void printPandaLinkState(byte[] bytes) {
    try {
      logger.debug("Got a joint message, parsing ...");
      PandaLinkState pls = PandaLinkState.parseFrom(bytes);
      PandaLinkState.Position tmpPosition = pls.getPos();
      PandaLinkState.Orientation tmpOrientation = pls.getOrient();
      PandaLinkState.TwistLinear tmpTwistLinear = pls.getTl();
      PandaLinkState.TwistAngular tmpTwistAngular = pls.getTa();
      // panda::panda_link0: pos(-3.0621333E-8,-1.5197388E-8,3.3411725E-5), orient(0.0,0.0,0.0,0.0), twist-linear(0.0,0.0,0.0), twist-angular(0.0,0.0,0.0)

      logger.printf(Level.INFO,topicPattern + ": pos(% .5f,% .5f,% .5f), ori(%.1f,%.1f,%.1f,%.1f)," +
              " twL(%.1f,%.1f,%.1f), twA(%.1f,%.1f,%.1f)",
          pls.getName(),
          tmpPosition.getPositionX(),
          tmpPosition.getPositionY(),
          tmpPosition.getPositionZ(),
          tmpOrientation.getOrientationX(),
          tmpOrientation.getOrientationY(),
          tmpOrientation.getOrientationZ(),
          tmpOrientation.getOrientationW(),
          tmpTwistLinear.getTwistLinearX(),
          tmpTwistLinear.getTwistLinearY(),
          tmpTwistLinear.getTwistLinearZ(),
          tmpTwistAngular.getTwistAngularX(),
          tmpTwistAngular.getTwistAngularY(),
          tmpTwistAngular.getTwistAngularZ());
    } catch (InvalidProtocolBufferException e) {
      e.printStackTrace();
    }
  }

  private void printRobotConfig(byte[] bytes) {
    try {
      logger.debug("Got a robotConfig message, parsing ...");
      RobotConfig robotConfig = RobotConfig.parseFrom(bytes);
      logger.info("robotConfig: speed = {}, loopTrajectory = {}, planningMode = {}",
          robotConfig.getSpeed(),
          robotConfig.getLoopTrajectory(),
          robotConfig.getPlanningMode().toString());
    } catch (InvalidProtocolBufferException e) {
      e.printStackTrace();
    }
  }

  private void printDataConfig(byte[] bytes) {
    try {
      logger.debug("Got a dataConfig message, parsing ...");
      DataConfig dataConfig = DataConfig.parseFrom(bytes);
      logger.info("dataConfig: position = {}, orientation = {}, twistLinear = {}, twistAngular = {}, publishRate = {}",
          dataConfig.getEnablePosition(),
          dataConfig.getEnableOrientation(),
          dataConfig.getEnableTwistLinear(),
          dataConfig.getEnableTwistAngular(),
          dataConfig.getPublishRate());
    } catch (InvalidProtocolBufferException e) {
      e.printStackTrace();
    }
  }

  private void printTrajectory(byte[] bytes) {
    try {
      logger.debug("Got a trajectory message, parsing ...");
      if (logger.isInfoEnabled()) {
        Trajectory trajectory = Trajectory.parseFrom(bytes);
        StringBuilder sb = new StringBuilder("trajectory:");
        for (Trajectory.Position pos : trajectory.getPosList()) {
          sb.append(" pos(")
              .append(String.format("% .5f,", pos.getX()))
              .append(String.format("% .5f,", pos.getY()))
              .append(String.format("% .5f", pos.getZ()))
              .append(")");
        }
        logger.info(sb.toString());
      }
    } catch (InvalidProtocolBufferException e) {
      e.printStackTrace();
    }
  }

  private void printNextStep(byte[] bytes) {
    logger.info("nextStep: {}",
        java.nio.ByteBuffer.wrap(bytes).getInt());
  }

}
